#ifndef __DEFINES_H__
#define __DEFINES_H__
#include <stdint.h>
#include <sys/types.h>
#include <math.h>
#include <alsa/asoundlib.h>
#include <asm/elphel/c313a.h>
#include <asm/elphel/exifa.h>
#include <ogg/ogg.h>
#include <libmkv.h>

#include "cammkv_cmd.h"
#include "cammkv_mov.h"
#include "cammkv_kml.h"

#define CAMMKV_FRAME_NOT_READY 1 // frame pointer valid, but not yet acquired
#define CAMMKV_FRAME_INVALID   2 // invalid frame pointer
#define CAMMKV_FRAME_CHANGED   3 // frame parameters have changed
#define CAMMKV_FRAME_NEXTFILE  4 // need to switch to a new file segment
#define CAMMKV_FRAME_BROKEN    5 // frame broken (buffer overrun)
#define CAMMKV_FRAME_FILE_ERR  6 // error with file I/O
#define CAMMKV_FRAME_MALLOC    7 // can not allocate memory
#define CAMMKV_FRAME_OTHER     8 // other errors

#define CAMMKV_FORMAT_NONE     0 // no video output
#define CAMMKV_FORMAT_OGM      1 // output as Ogg Media file
#define CAMMKV_FORMAT_JPEG     2 // output as individual JPEG files
#define CAMMKV_FORMAT_MOV      3 // output as Apple Quicktime
#define CAMMKV_FORMAT_MKV      4 // output as Matroska


#define MAP_OPTIONS MAP_FILE|MAP_PRIVATE

#define TIMESCALE		1000000
#define VCODEC_FOURCC   	"V_MS/VFW/FOURCC"

#define TRAILER_SIZE   		0x02
#define COMMAND_LOOP_DELAY 	500000 //0.5sec

#define DEFAULT_DURATION        1800 		/* !default duration (seconds) */
#define DEFAULT_LENGTH          1000000000 	/* !default segment length (B) */

//!Next 2 for Quicktime (mostly)
#define DEFAULT_FRAMES           16384 		/* Maximal number of frames in file segment 
					  		(each need 4* (1 + 1/frames_per_chunk) bytes for the frame index */
#define DEFAULT_FRAMES_PER_CHUNK 10 		/*second sparse index - for Quicktime fast forward */



// HEADER_SIZE is defined to be larger than actual header (later - with EXIF)  
// to use compile-time buffer
#define JPEG_HEADER_MAXSIZE    	0x300 // will not change

#define DEFAULT_AUDIO_FORMAT	SND_PCM_FORMAT_S16_LE
#define DEFAULT_AUDIO_CHANNELS 	2
#define DEFAULT_AUDIO_RATE	48000
#define AUDIO_RATE_MAX  	48000
#define AUDIO_RATE_MIN  	11025
#define AUDIO_CHANNELS_MAX  	2
#define AUDIO_CHANNELS_MIN  	1
#define AUDIO_DEF_VOLUME	"58981"

// BitMap info header for describe JPEG in MKV container
typedef struct
{
  uint32_t  bi_size;
  uint32_t  bi_width;
  uint32_t  bi_height;
  uint16_t  bi_planes;
  uint16_t  bi_bit_count;
  uint32_t  bi_compression;
  uint32_t  bi_size_image;
  uint32_t  bi_x_pels_per_meter;
  uint32_t  bi_y_pels_per_meter;
  uint32_t  bi_clr_used;
  uint32_t  bi_clr_important;
} alBITMAPINFOHEADER;


typedef struct {
	char                  debug_name[256];
	char                  path_prefix[256];
	char                  path[300];
  	int                   save_gp; 		//if non zero, current circbuf pointer will be saved to global pointer, so imgsrv can report /pointers

struct interframe_params_t    frame_params;
struct interframe_params_t    this_frame_params;
struct interframe_params_t    prev_frame_params;    // used in MOV format for recalculate FPS if need
	

	int                   fd_circ;   	//! file descriptor for circbuf
	int                   circ_buff_size;
	int                   cirbuf_rp; 	//!-1 - invalid
	int                   metadata_start;
	int                   frame_period ; 	//!in microseconds (need *1000 for store to MKV))

	int                   fd_head;   	//! file descriptor for JPEG header
	int                   head_size; 	//! JPEG header size
	char                  jpegHeader[JPEG_HEADER_MAXSIZE];

	elph_packet_chunk     packetchunks[7];
  	int                   chunk_index;
	int                   jpeg_len, 
			      bufSize, 
			      jpeg_len_last;    // for resizing buffer if jpeg size changed
	char		      *jpegBuffer;


	int                   width;
	int                   height;

	int		      video_format; 	// MOV,MKV,OGG,JPEG ...
	int		      set_format;

	int                   frameno;
	double                timescale; 	//! current timescale, default 1.0 
	double                set_timescale;
	double		      fps;

	int                   fd_fparmsall;   	//! file descriptor for sensor/compressor parameters

	int                   segment_duration;
	int                   segment_length,
			      segment_length_s;
	int                   buf_overruns;
	int                   buf_min;
	int                   max_frames;
	int                   set_max_frames;
  	int                   set_frames_skip;  //! will be copied to frames_skip if stopped or at start
  	int                   frames_skip;     	//! number of frames to skip after the one recorded (for time lapse)
        	                                //! if negetive - -(interval between frames in seconds)
  	int                   frames_skip_left; //! number of frames left to skip before the next one to be processed
                                          	//! if (frames_skip <0) - next timestamp to save an image
	// EXIF stuff
	int                   fd_exif;   	//! file descriptor for Exif data
	int                   exif; 		// 1 - calculate and include Exif headers in each frame
	int                   exifSize; 	//signed
  	unsigned char         ed[MAX_EXIF_SIZE];

	// Matroska 
        mk_Track	      *vtrack, 		// descriptors for store tracks(audio,video, GPS ??? and etc ) to container
			      *atrack;
    	mk_Writer 	      *mkv_file; 	// descriptor for MKV container
	alBITMAPINFOHEADER    vpic;
	int		      chapter_no;
	
	// Apple Quicktime
	int                   ivf; 		//! video file (jpeg, mov - open)
 	int		      *frame_lengths;
  	off_t                 frame_data_start; //! Quicktime (and else?) - frame data start (0xff 0xd8...)
 	int                   frames_per_chunk;
 	int                   set_frames_per_chunk;

  	//quicktime_t	      *qtfile;
	


	//kml stuff
	volatile int          kml_enable;       //! enable KML file generation
	int                   kml_used;         //! KML file generation used (change only when stopped)
	char                  kml_path[300];    //! full path for KML file (if any)
	FILE*                 kml_file;         //! stream to write kml file
	double                kml_horHalfFov;   //! half horizontal Fov (degrees)
	double                kml_vertHalfFov;  //! half vertical Fov (degrees)
	double                kml_near;         //! Use in KML "near" parameter (<=0 - don't use it)
	int                   kml_height_mode;  //! 1 - actual, 0 - ground
	double                kml_height;       //! extra height to add
	int                   kml_period;       //! generate PhotoOverlay for each kml_period seconds;
	int                   kml_last_ts;      //! last generated kml file timestamp
	int                   kml_last_uts;     //! last generated kml file timestamp, microseconds
      struct exif_dir_table_t kml_exif[ExifKmlNumber] ;  //! store locations of the fields needed for KML generations in the Exif block
	
	// audio stuff
	char 		      pcm_name[32];
	int 		      mmap_flag;		// method for access to sound card  1-mmap / 0-
	unsigned int	      period_time;		// audio chunk in usec, by default 20000 (20 ms)
	unsigned int	      buffer_time;		// buffer size for store audio chunks, in usec, by default 5 sec
	uint8_t 	      *audiobuf;
	snd_pcm_t 	      *capture_handle;
        snd_pcm_uframes_t     chunk_size;		// size audio chunk in frames
	size_t                chunk_bytes;
	int 		      audio_volume;
	snd_pcm_format_t      format;
	unsigned int          channels;
	unsigned int          rate;
	unsigned int	      bitDepth;
//	int		      corr_time,
//			      corr_times;
	uint64_t	      atime_start;		// FPGA time starting audio card, used for find videoframe for syncing
							// in MOV container
	 
	volatile int	      synced;
	volatile int	      a_stopping;
	volatile int	      v_stopping;
	volatile int	      audio_enabled;
	volatile int   	      can_capture;   	//audio subsystem flag
	volatile int 	      running;	
	volatile int	      starting;	
	volatile int	      done;		// for all threads, when done=1 - threads go to end

} _state_t;

#if 0
//////////////////////////////////////////////////////////////////////////
void lqt_video_append_timestamp(quicktime_t * file, int track,
		                                int64_t time, int duration);
void lqt_write_frame_header(quicktime_t * file, int track,
		int pic_num1,
		int64_t pic_pts, int keyframe);
int quicktime_write_data(quicktime_t *file, const uint8_t *data, int size);
void lqt_write_frame_footer(quicktime_t * file, int track);
void lqt_flush_timecode(quicktime_t * file, int track, int64_t time, int force);

//////////////////////////////////////////////////////////////////////////
#endif
extern int 		debug_level;
extern FILE* 		debug_file;

#define DBG(lvl, x...) \
    {if(debug_file && (debug_level>=lvl))\
    {fprintf(debug_file,"%s:%d:",__FILE__,__LINE__);fprintf(debug_file, x);fflush(debug_file);}}


extern _state_t 		sstate;
extern _state_t 		*state;
extern unsigned long    	*globalPars; 	
extern struct framepars_t      	*framePars;

extern unsigned long 	*ccam_dma_buf;           /* mmapped array */

extern pthread_mutex_t thread_lock;
extern pthread_mutex_t mkv_lock;
extern pthread_mutex_t stopping_lock;
extern pthread_cond_t  cond;

void * thread_v(void *arg);
void * thread_a(void *arg);

int parse_cmd(FILE* npipe); 

// interface container independent functions
// for writing audio/video frames
// inited when changed outut format
int (*v_write_func)(void);
int (*a_write_func)(void);
int (*av_open_func)(void);
int (*av_close_func)(void);

// container specific function declaration for
// writing audio/video frames
int v_mkvwrite(void);
int a_mkvwrite(void);
int v_movwrite(void);
int a_movwrite(void);

// Matroska specific
int dvr_mkvopen(void);
int dvr_mkvclose(void);
int dvr_mkvreset(void);

// QuickTime specific
int dvr_movopen(void);
int dvr_movclose(void);
int dvr_movreset(void);

int dvr_reset(void);
int dvr_free(void);

int  cammkv_start(void);
void cammkv_stop(void);
int  cammkv_free(void);
void cammkv_status(char * fn, int xml); 
int  cammkv_reset(void);

int  init_audio(void);
void set_audio_pars(void);
void set_audio_state(char *args); 
void set_audio_format(char *args); 
void set_audio_volume(char *args); 
void set_audio_tc(char *args);
int  deinit_audio(void);
int  stopping_audio(void);

void set_frames_skip(int d); 
void set_format(int d); 
void  set_prefix (const char * p); 

int init_video(void); 
int deinit_video(void);
int stopping_video(void);

unsigned long getGPValue(unsigned long GPNumber);
void get_fpga_time(unsigned long *rtc); 

#endif
