/*!***************************************************************************
*! FILE NAME  : cammkv_cmd.c
*! DESCRIPTION: Functions for parsing commands from pipe and interpretate it.
*! Copyright (C) 2010 Elphel, Inc.
*! -----------------------------------------------------------------------------**
*!  This program is free software: you can redistribute it and/or modify
*!  it under the terms of the GNU General Public License as published by
*!  the Free Software Foundation, either version 3 of the License, or
*!  (at your option) any later version.
*!
*!  This program is distributed in the hope that it will be useful,
*!  but WITHOUT ANY WARRANTY; without even the implied warranty of
*!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*!  GNU General Public License for more details.
*!
*!  You should have received a copy of the GNU General Public License
*!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*! -----------------------------------------------------------------------------**
*!
*!  $Log$
*/
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <inttypes.h>
#include <pthread.h>
#include <string.h>
#include <sys/time.h>

#include "config.h"
#include "defines.h"
#include "cammkv_jpeg.h"


/////////////////////////////////////////////////////////////////////////////////////////
static char 		cmdbuf[1024];
static int  		cmdbufp=0; 	// current input pointer in the command buffer (read from pipe)
static int  		cmdstrt=0; 	// start of the next partial command

void cammkv_set_debug_out(const char * fname) 
{
	int none=1;
	if (fname && strlen(fname) && strcmp(fname, "none") && strcmp(fname, "null")  && strcmp(fname, "/dev/null")) none=0;
	if (debug_file){
		if (strcmp(state->debug_name, "stdout") && strcmp(state->debug_name, "stderr")) fclose (debug_file);
		debug_file=NULL;
		state->debug_name[0]='\0';
	}
	if (!none) {
		if      (strcmp(fname, "stdout") ==0) debug_file=stdout;
		else if (strcmp(fname, "stderr") ==0) debug_file=stderr;
		else                                  debug_file=fopen(fname,"w+");
	}
	if (debug_file) {
		strncpy(state->debug_name,fname,sizeof(state->debug_name)-1);
		state->debug_name[sizeof(state->debug_name)-1]='\0';
	}

return;
}

//! needed for Quicktime - maybe something else?
void  cammkv_set_max_frames(int d) 
{
   state->set_max_frames=  d;
   if ((state->running==0) && (state->starting==0)) state->max_frames=  d;
}
void  cammkv_set_frames_per_chunk(int d) 
{
   state->set_frames_per_chunk=  d;
   if ((state->running==0) && (state->starting==0)) state->frames_per_chunk=  d;
}

void  cammkv_set_frames_skip(int d) 
{ 	// set frames to skip (for time lapse)
	state->set_frames_skip=  d;
	//if ((state->running==0) && (state->starting==0)) {
	if ((state->starting==0)) {
		state->frames_skip=state->set_frames_skip;
		state->frames_skip_left= 0;
	}
}

int cammkv_reset(void)
{
	state->buf_overruns=-1;
	state->synced = 0;
	cammkv_set_frames_skip(0); // don't skip after reset
return 0;
}

void  cammkv_set_prefix (const char * p) 
{
  strncpy(state->path_prefix, p, sizeof(state->path_prefix)-1);
  state->path_prefix[sizeof(state->path_prefix)-1]='\0';
}

void  cammkv_set_segment_duration(int sd) 	{ state->segment_duration=sd;}
void  cammkv_set_segment_length(int sl)   	{ state->segment_length= sl;}
void  cammkv_set_save_gp(int d) 		{state->save_gp=   d;}
void  cammkv_set_exif(int d) 			{ state->exif=   d;}

void  cammkv_set_timescale(double d) 
{ 
	state->set_timescale=  d;
	if ((state->running==0) && (state->starting==0)) {
		state->timescale=state->set_timescale;
	}
}


///kml stuff
void  cammkv_kml_set_enable(int d) 		{ state->kml_enable=d;}
void  cammkv_kml_set_horHalfFov (double dd) 	{ state->kml_horHalfFov=dd;}
void  cammkv_kml_set_vertHalfFov(double dd) 	{ state->kml_vertHalfFov=dd;}
void  cammkv_kml_set_height_mode(int d) 	{ state->kml_height_mode=d;}
void  cammkv_kml_set_height(double dd) 		{ state->kml_height=dd;}
void  cammkv_kml_set_period(int d) 		{ state->kml_period=d; state->kml_last_ts=state->kml_last_uts=0;}
void  cammkv_kml_set_near(double dd) 		{ state->kml_near=dd;}



void cammkv_set_audio_state(char *args) 
{
	if(strcmp(args, "on") == 0 || 
	   strcmp(args, "enable") == 0 ||
           strcmp(args, "1") == 0) {
		state->audio_enabled = 1;
	} else {
		state->audio_enabled = 0;
	}
	state->can_capture   = 0;
	DBG(3,"Request for audio state: %s\n", args);
}

void cammkv_set_audio_format(char *args) 
{
	char buf[8];
	int i, channels = -1, rate = -1;
	char *right = strstr(args, "/");
	if(right != NULL) {
		channels = atol(&right[1]);
		if(channels < AUDIO_CHANNELS_MIN) channels = AUDIO_CHANNELS_MIN;
		if(channels > AUDIO_CHANNELS_MAX) channels = AUDIO_CHANNELS_MAX;
		state->channels = channels;
	}
	for(i = 0; i < 7; i++) {
		if(args[i] != '\0' && args[i] != '/')
			buf[i] = args[i];
		else break;
	}
	buf[i] = '\0';
	if(buf[0] != '\0') {
		rate = atol(buf);
		if(rate < AUDIO_RATE_MIN) rate = AUDIO_RATE_MIN;
		if(rate > AUDIO_RATE_MAX) rate = AUDIO_RATE_MAX;
		state->rate = rate;
	}
	DBG(3,"Request for audio format: %s; after apply: rate == %d; channels == %d\n", args, state->rate, state->channels);
}

void cammkv_set_audio_volume(char *args) 
{
	int vol = atoi(args);
	if(vol > 65535) vol = 65535;
	if(vol < 0) 	vol = 0;
	state->audio_volume = vol;
}



void cammkv_set_format(int d) 
{
	if ((state->running==0) && (state->starting==0)) {
		state->set_format=d;
		state->video_format=d;
		switch(d) {
			case CAMMKV_FORMAT_NONE:
			case CAMMKV_FORMAT_JPEG:
				v_write_func=cammkv_frame_jpeg;
				a_write_func =cammkv_end_jpeg;// dummy - nothing doing	
				av_open_func=cammkv_start_jpeg;
				av_close_func=cammkv_end_jpeg;	
				state->audio_enabled=0;
				state->synced=0;
				break;
			case CAMMKV_FORMAT_OGM:
				DBG(0, "Video output to OGG/OGM not released yet.\n");
				break;
			case CAMMKV_FORMAT_MOV:
				v_write_func = v_movwrite;
				a_write_func = a_movwrite;
				av_open_func = dvr_movopen;
				av_close_func= dvr_movclose;
				state->synced=0;
				break;
			case CAMMKV_FORMAT_MKV:
			default:
				v_write_func = v_mkvwrite;
				a_write_func = a_mkvwrite;
				av_open_func = dvr_mkvopen;
				av_close_func= dvr_mkvclose;
				state->synced=0;
		}
	}	
	
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// will read from pipe, return pointer to null terminated string if available, NULL otherwise
static char * getLineFromPipe(FILE* npipe) 
{
	int fl;
	char * nlp;
	//!remove used string if any
	if (cmdstrt > 0) {
		//!moving overlapping strings
		memmove(cmdbuf, &cmdbuf[cmdstrt], sizeof(cmdbuf)- cmdstrt);
		cmdbufp-=cmdstrt;
		cmdstrt=0;
	}
	//! is there any complete string in a buffer?
	if (!cmdbufp) cmdbuf[cmdbufp]=0; //!null-terminate first access (probably not needed for the static buffer
	nlp= strpbrk(cmdbuf,";\n");
	if (!nlp) { //!no complete string, try to read more
		fl=fread(&cmdbuf[cmdbufp], 1,sizeof(cmdbuf)-cmdbufp-1,npipe);
		cmdbuf[cmdbufp+fl]=0;
		//! is there any complete string in a buffer after reading?
		nlp= strpbrk(&cmdbuf[cmdbufp],";\n"); //! there were no new lines before cmdbufp
		cmdbufp+=fl; //!advance pointer after pipe read
	}
	if (nlp) {
		nlp[0]=0;
		cmdstrt=nlp-cmdbuf+1;
		for (fl=0; cmdbuf[fl] && strchr(" \t",cmdbuf[fl]); fl++);
		return &cmdbuf[fl];
	}
return NULL;
}

// command[= \t]*args[ \t]*
int parse_cmd(FILE* npipe) 
{
	char *cmd, *args, *argse;
	int d;
	double dd;

	//!skip empty commands
	while(((cmd=getLineFromPipe(npipe))) && !cmd[0]) ;
	if (!cmd) return 0; //! nothing in the pipe
	DBG(3,"cmd[0]=%d:%s\n",(int) cmd[0],cmd);
	args=strpbrk(cmd,"= \t");
	//! is it just a single word command or does it have parameters?
	if (args) {
		args[0]=0;
		args++;
		while (strchr("= \t",args[0])) args++;
		if (args[0]) {
			//! ltrim (args)
			for (argse=strchr(args,'\0')-1; strchr("= \t",argse[0]);argse--) argse[0]='\0';
		}
		if (!args[0]) args=NULL;
	}
	//! now cmd is trimmed, arg is NULL or a pointer to trimmed command arguments
	if      (strcmp(cmd, "start")==0) {
		if(state->running==0) state->starting=1;
	} else if (strcmp(cmd, "reset")==0) { //! will reset pointer to the last acquired frame (if any)
		cammkv_reset();
	} else if (strcmp(cmd, "stop")==0) {
		cammkv_stop();
		cammkv_free();
	} else if (strcmp(cmd, "exit")==0) { 
		cammkv_stop();
		cammkv_free();
		state->done=1;
		pthread_mutex_unlock (&thread_lock);
	} else if (strcmp(cmd, "status")==0) {
		cammkv_status(args, 0); 
	} else if (strcmp(cmd, "xstatus")==0) {
		cammkv_status(args, 1); 
	} else if (strcmp(cmd, "save_gp")==0) {
		if ((args) && (((d= strtol(args, NULL, 10)))>=0)) cammkv_set_save_gp(d);
	} else if (strcmp(cmd, "exif")==0) {
		if ((args) && (((d= strtol(args, NULL, 10)))>=0)) cammkv_set_exif(d);
	} else if (strcmp(cmd, "debug")==0) {
		cammkv_set_debug_out(args);
	} else if (strcmp(cmd, "debuglev")==0) {
		d= strtol(args, NULL, 10);
		debug_level=(d?d:0);
	} else if (strcmp(cmd, "duration")==0) {
		if (!(args) || (((d= strtol(args, NULL, 10)))<=0)) d=DEFAULT_DURATION;
   		cammkv_set_segment_duration(d);
	} else if (strcmp(cmd, "length")==0) {
		if (!(args) || (((d= strtol(args, NULL, 10)))<=0)) d=DEFAULT_LENGTH;
   		cammkv_set_segment_length(d);
	} else if (strcmp(cmd, "prefix")==0) {
		if (args) cammkv_set_prefix (args);
	} else if (strcmp(cmd, "timescale")==0) {
		dd= strtod(args,NULL);
		cammkv_set_timescale((dd?dd:1.0));
		//!TODO: fix period calculation/check for frame skipping (just disable in frame skip mode?)
		//!TODO: add time period (system clock), not just frame skipping
	} else if (strcmp(cmd, "frameskip")==0) {
		d= strtol(args, NULL, 10);
		cammkv_set_frames_skip(d);
	} else if (strcmp(cmd, "timelapse")==0) { //! period (in seconds) between stored frames
		d= strtol(args, NULL, 10);
		cammkv_set_frames_skip(-d);
	} else if (strcmp(cmd, "format")==0) {
		if (args) {
			if      (strcmp(args,  "none")==0) cammkv_set_format(0);
			else if ((strcmp(args, "ogm" )==0) || (strcmp(args, "ogg")==0)) cammkv_set_format(CAMMKV_FORMAT_OGM);
			else if ((strcmp(args, "jpeg")==0) || (strcmp(args, "jpg")==0)) cammkv_set_format(CAMMKV_FORMAT_JPEG);
			else if (strcmp(args,  "mov" )==0)                              cammkv_set_format(CAMMKV_FORMAT_MOV);
			else if (strcmp(args,  "mkv" )==0) 				cammkv_set_format(CAMMKV_FORMAT_MKV);
		} else cammkv_set_format(CAMMKV_FORMAT_MKV);
	//kml stuff ------------------------------------
	} else if (strcmp(cmd, "kml")==0) {
		if ((args) && (((d= strtol(args, NULL, 10)))>=0)) 
	   		cammkv_kml_set_enable(d);
	} else if (strcmp(cmd, "kml_hhf")==0) {
		dd= strtod(args,NULL);
	   	cammkv_kml_set_horHalfFov(dd);
	} else if (strcmp(cmd, "kml_vhf")==0) {
		dd= strtod(args,NULL);
	   	cammkv_kml_set_vertHalfFov(dd);
	} else if (strcmp(cmd, "kml_near")==0) {
		dd= strtod(args,NULL);
	   	cammkv_kml_set_near(dd); // distance to PhotoOverlay
	} else if (strcmp(cmd, "kml_alt")==0) {
		if (args) {
			if      (strcmp(args, "gps"   )==0) cammkv_kml_set_height_mode(1);
			else if (strcmp(args, "ground")==0) cammkv_kml_set_height_mode(0);
		}
	} else if (strcmp(cmd, "kml_height")==0) {
		dd= strtod(args,NULL);
	   	cammkv_kml_set_height(dd);
	} else if (strcmp(cmd, "kml_period")==0) {
		d= strtol(args, NULL, 10);
		cammkv_kml_set_period((d?d:1));
	// --- Audio stuff ------------------------------------
	} else if (strcmp(cmd, "audio")==0) {
		if (args) {
			cammkv_set_audio_state(args);
		}
	} else if (strcmp(cmd, "audio_fmt")==0) {
		if (args) {
			cammkv_set_audio_format(args);
		}
	} else if (strcmp(cmd, "audio_vol")==0) {
		if (args) {
			cammkv_set_audio_volume(args);
		}
	} else 
		return -1;
  
return 1;
}

