/*!***************************************************************************
*! FILE NAME  : video_thr.c
*! DESCRIPTION: It is implementation of thread, which captured video data 
*! and dump it to audio track Matroska container.
*! Copyright (C) 2010 Elphel, Inc.
*! -----------------------------------------------------------------------------**
*!  This program is free software: you can redistribute it and/or modify
*!  it under the terms of the GNU General Public License as published by
*!  the Free Software Foundation, either version 3 of the License, or
*!  (at your option) any later version.
*!
*!  This program is distributed in the hope that it will be useful,
*!  but WITHOUT ANY WARRANTY; without even the implied warranty of
*!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*!  GNU General Public License for more details.
*!
*!  You should have received a copy of the GNU General Public License
*!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*! -----------------------------------------------------------------------------**
*!
*!  $Log$
*/
#define __USE_POSIX199506 
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <inttypes.h>
#include <pthread.h>
#include <string.h>
#include <sys/time.h>

#include <ogg/ogg.h>

#include "ogmstreams.h"
#include "cammkv_kml.h"
#include "config.h"
#include "defines.h"

char trailer[TRAILER_SIZE] = {0xff,0xd9};

int  getImageFrame (void) 
{
	int  	offset;
	void 	*ptr;
	int 	timestamp_start;
	int 	rslt,diffs;
	int 	*ifp_this = (int *)&(state->this_frame_params) ;
	unsigned char frame_packet_type = PACKET_IS_SYNCPOINT;

	DBG(5,"GetImageFrame starting ...\n");

	//! This is probably needed only for Quicktime (not to exceed already allocated frame index)
	if (state->frameno >= (state->max_frames)) {
		DBG(3,"FRAME_CHANGED: frameno %d max_frames %d\n",state->frameno, state->max_frames);
		return -CAMMKV_FRAME_CHANGED;
	}
	//! Format changed?
	if (state->video_format != state->set_format) {
		DBG(3,"Video format changed\n");
		return -CAMMKV_FRAME_CHANGED;
	}
	//!   check if file size is exceeded (assuming fopen),-CAMMKV_FRAME_CHANGED will trigger a new segment
	//if ((state->vf) && (state->segment_length >=0) && (ftell(state->vf) > state->segment_length)) return -CAMMKV_FRAME_CHANGED;
	//!same for open
	if (((state->ivf)>=0) && (state->segment_length >=0) && (lseek(state->ivf, 0, SEEK_CUR) > state->segment_length)){ 
			DBG(3,"Segment MOV length\n");
			return -CAMMKV_FRAME_CHANGED;
	}

	//! check the frame pointer is valid
	if (lseek(state->fd_circ, state->cirbuf_rp, SEEK_SET) <0) 
		return -CAMMKV_FRAME_INVALID; 				//!it will probably be that allready
	//! is the frame ready?
	if (lseek(state->fd_circ, LSEEK_CIRC_READY, SEEK_END) <0) 
		return -CAMMKV_FRAME_NOT_READY; 			//! frame pointer valid, but no frames yet

	// process skipping frames. TODO: add - skipping time between frames 
	// (or better -  actual time period - use the nearest frame) instead of the frame number
	if ( (state->frames_skip > 0) && (state->frames_skip_left > 0 )) { //!skipping frames, not seconds.
		state->cirbuf_rp=lseek(state->fd_circ,LSEEK_CIRC_NEXT,SEEK_END);
		//!optionally save it to global read pointer (i.e. for debugging with imgsrv "/pointers")
		if (state->save_gp) 
			lseek(state->fd_circ,LSEEK_CIRC_SETP,SEEK_END);
		state->frames_skip_left--;
		return -CAMMKV_FRAME_NOT_READY; //! the required frame is not ready
	}

	//! Get metadata
	state->metadata_start = state->cirbuf_rp - 32;
	if (state->metadata_start < 0) 
		state->metadata_start+=state->circ_buff_size;
	memcpy (&(state->this_frame_params), (unsigned long * ) &ccam_dma_buf[state->metadata_start>>2],32);
	state->jpeg_len=state->this_frame_params.frame_length; //! frame_params.frame_length are now the length of bitstream
	if (state->this_frame_params.signffff !=0xffff) {
		DBG(0, "%s:%d: wrong signature - %d\r\n",__FILE__,__LINE__,(int) state->this_frame_params.signffff);
		DBG(1, "state->cirbuf_rp=0x%x\r\n",(int) state->cirbuf_rp);
		DBG(1, "%08x %08x %08x %08x %08x %08x %08x %08x\r\n",\
			ifp_this[0],ifp_this[1],ifp_this[2],ifp_this[3],ifp_this[4],ifp_this[5],ifp_this[6],ifp_this[7]);
		return -CAMMKV_FRAME_BROKEN;
	}
	if(state->video_format == CAMMKV_FORMAT_MKV) {
		DBG(3,"jpeg_len=%d jpeg_len_last=%d\n",state->jpeg_len,state->jpeg_len_last);
		if(state->jpeg_len > state->jpeg_len_last) {
			state->jpeg_len_last = state->jpeg_len;
			if((ptr=realloc(state->jpegBuffer, 
				state->head_size+state->jpeg_len+TRAILER_SIZE+state->exifSize)) != NULL) {
				DBG(3,"reallocated buffer: h=%d l=%d t=%d e=%d\n",state->head_size,state->jpeg_len, TRAILER_SIZE,state->exifSize);
				state->jpegBuffer=ptr;
			}
			else {
				DBG(0,"Can't realloc memory for JPEG buffer, size %d!\n",state->head_size+state->jpeg_len+TRAILER_SIZE);
				return -CAMMKV_FRAME_BROKEN;
			}
		}
	}
	// find correct video frame by for syncing A/V
	// if audio enabled - waiting while started audio 
		if(state->audio_enabled) {
			if(!state->can_capture)
				return -CAMMKV_FRAME_NOT_READY; 	// not started audio, can't sync streams
		}
	// audio thread starting later video threading, then syncing video(skip some frames) while timestamp 
        // videoframe will be near timestamp audio started and  dumping data together
	while(1) {
		//!   find location of the timestamp and copy it to the frame_params structure
		timestamp_start= state->cirbuf_rp+((state->jpeg_len+CCAM_MMAP_META+3) & (~0x1f)) + 32 - CCAM_MMAP_META_SEC; //! magic shift - should index first byte of the time stamp
		if (timestamp_start >= state->circ_buff_size) timestamp_start-=state->circ_buff_size;
		memcpy (&(state->this_frame_params.timestamp_sec), (unsigned long * ) &ccam_dma_buf[timestamp_start>>2],8);
		if(!state->audio_enabled)
			break;
		if(state->synced)
			break;
		diffs=state->atime_start-(state->this_frame_params.timestamp_sec*1000000+state->this_frame_params.timestamp_usec);
		DBG(3,"Diffs A/V: %d\n",diffs);
	        if(diffs < 20000) {
			memcpy (&(state->frame_params.timestamp_sec), &(state->this_frame_params.timestamp_sec), 8); //store timestamp
			//memcpy (&(state->prev_frame_params.timestamp_sec), &(state->this_frame_params.timestamp_sec), 8); //store timestamp
			state->synced=1;
			break;
		}
		else {
			state->cirbuf_rp=lseek(state->fd_circ, LSEEK_CIRC_NEXT, SEEK_END);
			//!optionally save it to global read pointer (i.e. for debugging with imgsrv "/pointers")
			if (state->save_gp) lseek(state->fd_circ,LSEEK_CIRC_SETP,SEEK_END);
				return -CAMMKV_FRAME_NOT_READY; 			
		}
	}

	//! verify that the essential current frame params did not change, 
	//if they did - return an error (need new file header) 
	if ((state->frame_params.width != state->this_frame_params.width) || 
			(state->frame_params.height != state->this_frame_params.height)) { 
		DBG(3,"Parameters of frame changed!\n");
		return -CAMMKV_FRAME_CHANGED; // picture size changed
	}
	//!   check if file duration (in seconds) exceeded ,-CAMMKV_FRAME_NEXTFILE will trigger a new segment
	if ((state->segment_duration > 0) && 
			((state->this_frame_params.timestamp_sec-state->frame_params.timestamp_sec) > state->segment_duration)){
		state->frame_params.timestamp_sec=state->this_frame_params.timestamp_sec; 
		DBG(3,"Segment duration %d %ld\n",state->segment_duration, 
				state->this_frame_params.timestamp_sec-state->frame_params.timestamp_sec);
		return -CAMMKV_FRAME_NEXTFILE;
	}
	//! check if (in timelapse mode)  it is too early for the frame to be stored
	if ((state->frames_skip < 0) && (state->frames_skip_left > state->this_frame_params.timestamp_sec) ) {
		state->cirbuf_rp=lseek(state->fd_circ,LSEEK_CIRC_NEXT,SEEK_END);
		//!optionally save it to global read pointer (i.e. for debugging with imgsrv "/pointers")
		if (state->save_gp) lseek(state->fd_circ,LSEEK_CIRC_SETP,SEEK_END);
		return -CAMMKV_FRAME_NOT_READY; //! the required frame is not ready
	}

	if (state->exif) {
		//! update the Exif header with the current frame metadata
		//     updateExif(ep, state->ed, &(state->frame_params));
		state->exifSize=lseek(state->fd_exif, 1 ,SEEK_END); // at the beginning of page 1 - position == page length
		if (state->exifSize > 0) {
			//state->this_frame_params.meta_index
			lseek(state->fd_exif, state->this_frame_params.meta_index, SEEK_END); //! select meta page to use (matching frame)
			rslt=read (state->fd_exif, state->ed, state->exifSize);
			if (rslt<0) rslt=0;
			state->exifSize=rslt;
		} else state->exifSize=0;
	} else state->exifSize=0;
//////////////////////
//! prepare a packet to be sent (a lst of memory chunks)
#if 1
	state->chunk_index=0;
	state->packetchunks[state->chunk_index  ].bytes=1;
	state->packetchunks[state->chunk_index++].chunk=&frame_packet_type;
	if (state->exif>0) {//! insert Exif
		state->packetchunks[state->chunk_index  ].bytes=2;
		state->packetchunks[state->chunk_index++].chunk=state->jpegHeader;
		state->packetchunks[state->chunk_index  ].bytes=state->exifSize;
		state->packetchunks[state->chunk_index++].chunk= state->ed;
		state->packetchunks[state->chunk_index  ].bytes=state->head_size-2;
		state->packetchunks[state->chunk_index++].chunk= &(state->jpegHeader[2]);
	} else {
		state->packetchunks[state->chunk_index  ].bytes=state->head_size;
		state->packetchunks[state->chunk_index++].chunk= state->jpegHeader;
	}
	/*! JPEG image data may be split in two segments (rolled over buffer end) - process both variants */
	if ((state->cirbuf_rp + state->jpeg_len) > state->circ_buff_size) { //! two segments
		/*! copy from the beginning of the frame to the end of the buffer */
		state->packetchunks[state->chunk_index  ].bytes=state->circ_buff_size-state->cirbuf_rp;
		state->packetchunks[state->chunk_index++].chunk= (unsigned char*) &ccam_dma_buf[state->cirbuf_rp>>2];
		/*! copy from the beginning of the buffer to the end of the frame */
		state->packetchunks[state->chunk_index  ].bytes=state->jpeg_len - (state->circ_buff_size-state->cirbuf_rp);
		state->packetchunks[state->chunk_index++].chunk= (unsigned char*) &ccam_dma_buf[0];
	} else { // single segment
		/*! copy from the beginning of the frame to the end of the frame (no buffer rollovers) */
		//        sendBuffer((void *) &ccam_dma_buf[state->cirbuf_rp>>2], state->jpeg_len);
		state->packetchunks[state->chunk_index  ].bytes=state->jpeg_len;
		state->packetchunks[state->chunk_index++].chunk= (unsigned char*)  &ccam_dma_buf[state->cirbuf_rp>>2];
	}
	state->packetchunks[state->chunk_index  ].bytes=2;
	state->packetchunks[state->chunk_index++].chunk= (unsigned char*) trailer;

///////////////////
#else
	offset=0;
	if (state->exif>0) {//! insert Exif
		if(state->head_size+state->jpeg_len+TRAILER_SIZE+state->exifSize > state->head_size+state->jpeg_len+TRAILER_SIZE) {
			if((ptr=realloc(state->jpegBuffer, state->head_size+state->jpeg_len+TRAILER_SIZE+state->exifSize)) != NULL)
				state->jpegBuffer=ptr;
			else {
				DBG(0,"Can't realloc memory for JPEG buffer, size %d!\n",state->head_size+state->jpeg_len+TRAILER_SIZE+state->exifSize);
				return -CAMMKV_FRAME_BROKEN;
			}
		}
		memcpy(state->jpegBuffer, state->jpegHeader, 2);
		offset+=2;
		memcpy(state->jpegBuffer+offset, state->ed, state->exifSize);
		offset+=state->exifSize;
		memcpy(state->jpegBuffer+offset, state->jpegHeader+2, state->head_size-2);
		offset+=(state->head_size-2);
	} else {
		memcpy(state->jpegBuffer, state->jpegHeader, state->head_size);
		offset+=state->head_size;
	}
	/*! JPEG image data may be split in two segments (rolled over buffer end) - process both variants */
	if ((state->cirbuf_rp + state->jpeg_len) > state->circ_buff_size) { //! two segments
		/*! copy from the beginning of the frame to the end of the buffer */
		memcpy(state->jpegBuffer+offset, (void *) &ccam_dma_buf[state->cirbuf_rp>>2], 
				state->circ_buff_size-state->cirbuf_rp);
		offset+=(state->circ_buff_size-state->cirbuf_rp); 
		/*! copy from the beginning of the buffer to the end of the frame */
		memcpy(state->jpegBuffer+offset, (void *) &ccam_dma_buf[0], 
				state->jpeg_len - (state->circ_buff_size-state->cirbuf_rp)); 
		offset+=(state->jpeg_len - (state->circ_buff_size-state->cirbuf_rp));
	} else { // single segment
		/*! copy from the beginning of the frame to the end of the frame (no buffer rollovers) */
		memcpy(state->jpegBuffer+offset, (void *) &ccam_dma_buf[state->cirbuf_rp>>2], state->jpeg_len);
		offset+=state->jpeg_len; 
	}
	memcpy(state->jpegBuffer+offset,(void *)trailer, 2);
	offset+=2;
#endif
   	if (state->kml_used) 
		rslt=cammkv_frame_kml() ; // will turn on state->kml_used if it can

	//!advance frame pointer
	state->frameno++;
#if 0
	state->segment_length-=offset;  
        DBG(3,"Length=%d - %d \n", state->segment_length, state->segment_length_s);	
	if(state->segment_length < 0) {
		//state->segment_length=DEFAULT_LENGTH;     	
		state->segment_length=state->segment_length_s;     	
		return -CAMMKV_FRAME_NEXTFILE;
	}	
#endif
	state->cirbuf_rp=lseek(state->fd_circ, LSEEK_CIRC_NEXT, SEEK_END);
	//!optionally save it to global read pointer (i.e. for debugging with imgsrv "/pointers")
	if (state->save_gp) lseek(state->fd_circ,LSEEK_CIRC_SETP,SEEK_END);

	if (state->frames_skip > 0) {
		state->frames_skip_left= state->frames_skip;
	} else if (state->frames_skip < 0) {
		state->frames_skip_left+= -(state->frames_skip);
	}

return 0;
}

//////////////// Functions for dumping video frames to various containers //////////////////////////////////////////////////////

#if 0
int v_movwrite(void)
{
	int time_diff=
		(uint64_t)((state->this_frame_params.timestamp_usec - state->prev_frame_params.timestamp_usec)+
			1000000*(state->this_frame_params.timestamp_sec  - state->prev_frame_params.timestamp_sec));
	int result = 0;
	int track=0;
	double framerate=1000./((double)time_diff/1000.);

	DBG(4,"->>>>>frame time diff=%d fps= %6.2f\n",time_diff, framerate);
	if(state->fps != framerate) {
		state->fps=framerate;
		quicktime_set_framerate(state->qtfile, state->fps);
	}
	result= quicktime_write_frame(state->qtfile, state->jpegBuffer, state->head_size+state->jpeg_len+state->exifSize+TRAILER_SIZE,
					track);
        // for recalculate FPS and setting correct timestamps in MOV container for syncing A/V
	memcpy (&(state->prev_frame_params.timestamp_sec), &(state->this_frame_params.timestamp_sec), 8); //store timestamp

return result;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#endif

/** Thread for get videoframes 
 */
void * thread_v(void *arg)
{
	int 		rslt;

        pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	while(!(state->done)) {
		DBG(3,"Waiting to start capture video...\n");
		pthread_mutex_lock (&thread_lock);
		pthread_mutex_unlock (&thread_lock);
		if(state->done)
			break;
		DBG(3,"start capture video...state->v_stopping=%d state->running=%d\n",state->v_stopping,state->running);
		while(state->running) {
		       if(state->v_stopping)
		       		break;	       
			switch ((rslt=-getImageFrame ())) {
				case 0: //dump video data to container
				        DBG(5,"Writing video data...\n");	
					pthread_mutex_lock (&mkv_lock);
					v_write_func();	
					pthread_mutex_unlock (&mkv_lock);
					continue;	
					break;
				case  CAMMKV_FRAME_CHANGED:   //! frame parameters have changed(pic size - not FPS)
						// close current chunk file, and open MKV with new picture size
					DBG(3,"Frame changed\n");
				case  CAMMKV_FRAME_INVALID:   //! invalid frame pointer
					DBG(3,"Frame invalid\n");
				case  CAMMKV_FRAME_NEXTFILE:  //! next file needed (need to switch to a new file (time/size exceeded limit)
					DBG(3,"Frame next file\n");
				case  CAMMKV_FRAME_BROKEN:    //! frame broken (buffer overrun)
							   //! restart the file
					DBG(3,"Thread_V ...next_file \n");
					pthread_mutex_lock (&mkv_lock);
					av_close_func();
					state->frameno=0;
					memcpy (&(state->frame_params),&(state->this_frame_params), 32);
					av_open_func();
					pthread_mutex_unlock (&mkv_lock);
					continue;	
					break;
				case CAMMKV_FRAME_NOT_READY:        //!  just wait for the frame to appear at the current pointer
					DBG(4,"Thread_V waiting ...not_ready \n");
					if(lseek(state->fd_circ, LSEEK_CIRC_WAIT, SEEK_END)<0) {
						DBG(3,"Video waiting ... ERROR LSEEK_CIRC_WAIT \n");
						lseek(state->fd_circ, LSEEK_CIRC_NEXT, SEEK_END);
					}
					break;
				case  CAMMKV_FRAME_FILE_ERR:  //! error with file I/O
				case  CAMMKV_FRAME_OTHER:     //! other errors
					DBG(0,"%s:line %d - error=%d\n",__FILE__,__LINE__,rslt);
					break;
				default:
					DBG(0,"%s:line %d - should not get here (rslt=%d)\n",__FILE__,__LINE__,rslt);
					exit (-1);

			} //switch
			usleep(20);
		} // running

		pthread_mutex_lock (&stopping_lock); 
	        DBG(3,"Stopping video...\n");	
		
		pthread_mutex_lock (&mkv_lock);
		av_close_func();
		pthread_mutex_unlock (&mkv_lock);
		
		state->v_stopping=2;
		pthread_cond_signal(&cond);
	        DBG(3,"Stopped video...\n");	
		pthread_mutex_unlock (&stopping_lock);
		usleep(20);

	} //done

DBG(2,"End of video thread.\n");
return NULL;
}

int init_video(void) 
{
	int 	timestamp_start,rslt;
	int 	next_metadata_start, 
		next_jpeg_len, fp;
	// for debugging
  	int * ifp =      (int *) &(state->frame_params) ;
	int * ifp_this = (int *) &(state->this_frame_params) ;

  	if (state->kml_enable) cammkv_init_kml() ; // do nothing

	DBG(1,"Starting init video recording subsytem ...\n");

	state->frameno=0;
	state->timescale=state->set_timescale;//! current timescale, default 1
	state->max_frames=state->set_max_frames;
	state->running =0;
	state->starting=1; //!may be already set
	state->video_format=state->set_format;

	//! Check/set circbuf read pointer
	DBG(3,"1: state->cirbuf_rp=0x%x\n",state->cirbuf_rp);
	if ((state->cirbuf_rp <0) 
			|| (lseek(state->fd_circ, state->cirbuf_rp, SEEK_SET) < 0) 
			|| (lseek(state->fd_circ, LSEEK_CIRC_VALID, SEEK_END) < 0 )) {
		DBG(3,"2: state->cirbuf_rp=0x%x\n",state->cirbuf_rp);
		state->cirbuf_rp=lseek(state->fd_circ, LSEEK_CIRC_LAST, SEEK_END);
		if (((fp=lseek(state->fd_circ,LSEEK_CIRC_PREV,SEEK_END)))>=0) 
			state->cirbuf_rp=fp; //!try to have 2 frames available for fps
		state->buf_overruns++;
		//! file pointer here should match state->rp; so no need to do    lseek(state->fd_circ,state->cirbuf_rp,SEEK_SET);
		//    state->buf_min=lseek(state->fd_circ,LSEEK_CIRC_FREE,SEEK_END);
		//    lseek(state->fd_circ,state->cirbuf_rp,SEEK_SET); //! restore after moving (side effect)
		state->buf_min=getGPValue(G_FREECIRCBUF);
	} else {
		//      b_free=lseek(state->fd_circ,LSEEK_CIRC_FREE,SEEK_END);
		//      lseek(state->fd_circ,state->cirbuf_rp,SEEK_SET); //! restore after moving (side effect)
		if (state->buf_min > getGPValue(G_FREECIRCBUF)) state->buf_min=getGPValue(G_FREECIRCBUF);

	}
	DBG(3,"3: state->cirbuf_rp=0x%x\n",state->cirbuf_rp);
	DBG(3,"4:lseek(state->fd_circ,LSEEK_CIRC_READY,SEEK_END)=%d\n",(int) lseek(state->fd_circ,LSEEK_CIRC_READY,SEEK_END));

	//! is this frame ready?
	if (lseek(state->fd_circ, LSEEK_CIRC_READY, SEEK_END) <0) 
		return -CAMMKV_FRAME_NOT_READY; //! frame pointer valid, but no frames yet
	DBG(3,"5: state->cirbuf_rp=0x%x\n",state->cirbuf_rp);
	state->metadata_start=(state->cirbuf_rp)-32;
	if (state->metadata_start<0) 
		state->metadata_start+=state->circ_buff_size;
	///==================================
	DBG(3,"6: Get frame params\n");
	memcpy (&(state->frame_params), (unsigned long * ) &(ccam_dma_buf[state->metadata_start>>2]),32);
	state->jpeg_len=state->frame_params.frame_length; //! frame_params.frame_length are now the length of bitstream

	if (state->frame_params.signffff !=0xffff) {
		DBG(0, "%s:%d: wrong signature - %d\r\n",__FILE__,__LINE__,(int) state->frame_params.signffff);
		state->cirbuf_rp=-1;
		DBG(1, "state->cirbuf_rp=0x%x\r\n",(int) state->cirbuf_rp);
		DBG(1, "%08x %08x %08x %08x %08x %08x %08x %08x\r\n",ifp[0],ifp[1],ifp[2],ifp[3],ifp[4],ifp[5],ifp[6],ifp[7]);
		return -CAMMKV_FRAME_BROKEN;
	}
	//!   find location of the timestamp and copy it to the frame_params structure
	///==================================
	DBG(3,"7: Get timestamp\n");
	timestamp_start=(state->cirbuf_rp)+((state->jpeg_len+CCAM_MMAP_META+3) & (~0x1f)) + 32 - CCAM_MMAP_META_SEC; //! magic shift - should index first byte of the time stamp
	if (timestamp_start >= state->circ_buff_size) 
		timestamp_start-=state->circ_buff_size;
	memcpy (&(state->frame_params.timestamp_sec), (unsigned long * ) &ccam_dma_buf[timestamp_start>>2],8);

	//! see if next frame is available   
	if ((lseek(state->fd_circ, LSEEK_CIRC_NEXT, SEEK_END) < 0 ) ||
			//! is that next frame ready?
			(((fp=lseek(state->fd_circ,LSEEK_CIRC_READY,SEEK_END))) < 0)) {
		lseek(state->fd_circ,state->cirbuf_rp,SEEK_SET); //!just in case - restore pointer
		return -CAMMKV_FRAME_NOT_READY; //! frame pointer valid, but no frames yet
	}
	next_metadata_start=fp-32;
	if (next_metadata_start<0) 
		next_metadata_start+= state->circ_buff_size;
	DBG(3,"7: next_metadata start 0x%X\n",next_metadata_start);
	memcpy (&(state->this_frame_params), (unsigned long * ) &ccam_dma_buf[next_metadata_start>>2],32);
	next_jpeg_len=state->this_frame_params.frame_length; //! frame_params.frame_length are now the length of bitstream
	if (state->this_frame_params.signffff !=0xffff) { //! should not happen ever
		DBG(0, "%s:%d: wrong signature - %d\r\n",__FILE__,__LINE__,(int) state->this_frame_params.signffff);
		DBG(1, "fp=0x%x\r\n",(int) fp);
		DBG(1, "%08x %08x %08x %08x %08x %08x %08x %08x\r\n",ifp_this[0],ifp_this[1],ifp_this[2],ifp_this[3],ifp_this[4],ifp_this[5],ifp_this[6],ifp_this[7]);
		state->cirbuf_rp=-1;
		return -CAMMKV_FRAME_BROKEN;
	}
	//! find location of the timestamp and copy it to the frame_params structure
	timestamp_start=fp+((next_jpeg_len+CCAM_MMAP_META+3) & (~0x1f)) + 32 - CCAM_MMAP_META_SEC; //! magic shift - should index first byte of the time stamp
	if (timestamp_start >= state->circ_buff_size) timestamp_start-=state->circ_buff_size;
	memcpy (&(state->this_frame_params.timestamp_sec), (unsigned long * ) &ccam_dma_buf[timestamp_start>>2],8);
	//! verify that the essential current frame params did not change, if they did - return an error (need new file header) 
	if ((state->frame_params.width  != state->this_frame_params.width) ||
			(state->frame_params.height != state->this_frame_params.height)) {
		//! Advance frame pointer to the next (caller should try again)
		state->cirbuf_rp=fp;
		return -CAMMKV_FRAME_CHANGED; // no yet checking for the FPS
	}
	//! calcualte the frame period - time difference (in microseconds)
	DBG(3,"8: Calculate frame period\n");

	state->frame_period=(state->this_frame_params.timestamp_usec - state->frame_params.timestamp_usec)+
		1000000*(state->this_frame_params.timestamp_sec  - state->frame_params.timestamp_sec);

	//! correct for timelapse modes:
	state->frames_skip=state->set_frames_skip;
	if (state->frames_skip > 0) {
		state->frames_skip_left=0;
		state->frame_period *= (state->frames_skip+1);
		//    state->frames_skip_left= state->set_frames_skip;
	} else if (state->frames_skip < 0) {
		state->frame_period=-(state->frames_skip) ; //! actual frame period will fluctuate to the nearest frame acquired (free running)
		state->frames_skip_left=state->frame_params.timestamp_sec;
	}

	//   state->time_unit= (uint64_t) (((double) state-> frame_period) * ((double) 10.) / ((double) state-> timescale));
	state->width=state->frame_params.width;
	state->height=state->frame_params.height;

	//!read JPEG header - it should stay the same for the whole file (restart new file if any parameters changed)
	//!rebuild JPEG header:
	lseek(state->fd_head, state->cirbuf_rp+1, SEEK_END); //!+1 to avoid condition when jpeg_start==0. overloaded lseek will ignore 5 LSBs when SEEK_END
	state->head_size=lseek(state->fd_head, 0, SEEK_END); /// In 8.0 the header size might change for some jp4 modes 
	if (state->head_size > JPEG_HEADER_MAXSIZE) {
		DBG(0,"%s:%d: Too big JPEG header (%d > %d)",__FILE__,__LINE__,state->head_size, JPEG_HEADER_MAXSIZE );
		return -2;
	}
	//! and read it
	lseek(state->fd_head,0,0);
	read (state->fd_head, state->jpegHeader, state->head_size);
	//! Restore read pointer to the original (now there may be no frame ready there yet)
	lseek(state->fd_circ, state->cirbuf_rp, SEEK_SET);
	if(state->video_format == CAMMKV_FORMAT_MKV ) {
		state->jpeg_len_last=state->jpeg_len;
		DBG(3,"Alloc buffer for MKV data jpeg_len=%d\n",state->jpeg_len);
		state->jpegBuffer = malloc(state->head_size+state->jpeg_len+TRAILER_SIZE+state->exifSize);
		if(state->jpegBuffer == NULL) {
			DBG(0,"Can't allocate memory for video buffer!\n");
			return -1;
		}
	}
	rslt=0;
   	if (state->kml_enable) 
		rslt=cammkv_start_kml() ; // will turn on state->kml_used if it can
   	if (rslt) 
		return rslt;

	state->running=1;
	state->starting=0;

return 0;
}



int deinit_video(void)
{
	DBG(3,"Deinit video");
	state->running=0;
	state->starting = 0;
	if(state->jpegBuffer != NULL) {
		free(state->jpegBuffer);
		state->jpegBuffer = NULL;
	}
	state->jpeg_len=0;

return 0;	
}

int stopping_video(void)
{
	state->v_stopping=1;
	pthread_mutex_lock (&stopping_lock);
	while(state->v_stopping==1)
		pthread_cond_wait(&cond,&stopping_lock);
	state->v_stopping=0;
	state->running = 0;
	pthread_mutex_unlock (&stopping_lock);

return 0;
}

